import logging
import numpy as np
import pandas as pd
from typing import Union, List
from torch.utils.data import Dataset
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler, StandardScaler

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('Data')


class SimpleData(Dataset):
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __len__(self):
        return len(self.y)

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class MILData():
    def __init__(self, dataframe: pd.DataFrame,
                 time_column: str = None, drop_time_col: bool = True) -> None:
        self.df = dataframe
        self.parameters_list = dataframe.columns.values.tolist()
        self.param_index_dict = {param: idx for idx, param in enumerate(self.parameters_list)}
        self.index_param_dict = {idx: param for idx, param in self.param_index_dict.items()}
        self.time_column = time_column
        self.data_is_scaled = False

        if drop_time_col:
            self.df.drop(columns=[time_column], inplace=True)

    def MIL_processing(self, dataframe: pd.DataFrame = None,
                       test_size: float = 0.3,
                       val_size: float = 0,
                       anonamly_column: str = 'Anomaly',
                       flight_id_column: str = 'flight_id',
                       different_length_method: str = 'pad',
                       pad_value: Union[int, str] = 'max',
                       anomalies: Union[int, List[int]] = 1) -> None:

        assert test_size >= 0
        assert val_size >= 0

        if dataframe is None:
            dataframe = self.df.copy()

        if isinstance(anomalies, int):
            anomalies = [anomalies]

        self.anomaly_column = anonamly_column
        self.flight_id_column = flight_id_column

        # Assumes that normal flights have an index of 0
        nom_flights_count = int(dataframe[dataframe[anonamly_column] == 0][flight_id_column].unique().shape[0])
        adverse_flights_count = 0
        for idx_anomaly in anomalies:
            adverse_flights_count += int(dataframe[dataframe[anonamly_column] == idx_anomaly][flight_id_column].
                                         unique().shape[0])
        n_flights = nom_flights_count + adverse_flights_count
        logger.info(f"There are: {n_flights}, {nom_flights_count} nominal and {adverse_flights_count} adverse")

        flight_class = np.ones((n_flights, 1)) * 10
        self.flights_lengths = np.zeros((n_flights, 1))
        self.fligths_indices = np.ones((n_flights, 1)) * -1
        split_index = np.arange(0, n_flights, 1).reshape(-1, 1)

        # Create a list of flights
        def append_dataframe_of_individual_flights(counter: int = 0, new_list: List[pd.DataFrame] = []):
            max_fl_length = 0
            min_fl_length = 1e6
            nom_anomalies = [0] + anomalies
            logger.debug(f"Labels to loop through: {nom_anomalies}")
            # Fill all the nominal flights and anomalies
            for idx in nom_anomalies:
                list_ids_of_flights = dataframe[dataframe[anonamly_column] == idx][flight_id_column].\
                                      unique().astype(int)
                for fl_id in list_ids_of_flights:
                    # L x D shape
                    fl = dataframe[(dataframe[flight_id_column] == fl_id) & (dataframe[anonamly_column] == idx)]
                    length_of_current_fl = len(fl)
                    logger.debug(f'Label: {idx}. Flight idx: {fl_id} has length: {length_of_current_fl}')
                    self.flights_lengths[counter] = length_of_current_fl

                    if length_of_current_fl > max_fl_length:
                        max_fl_length = length_of_current_fl
                    if length_of_current_fl < min_fl_length:
                        min_fl_length = length_of_current_fl

                    flight_class[counter] = idx
                    self.fligths_indices = fl_id
                    new_list.append(fl)
                    counter += 1

            return new_list, counter, max_fl_length

        temp, counter, self.max_flight_length = append_dataframe_of_individual_flights()

        # TODO: Make this work with variable flight lengths
        # Stacking flights together
        unique_flight_lengths = np.unique(self.flights_lengths.flatten())
        if len(unique_flight_lengths) == 1:
            try:
                data = np.stack(temp, axis=0)  # shape [# flights, sample per flight, n_features]
            except ValueError as err:
                if err == "all input arrays must have the same shape":
                    logger.error(err)
                raise
            except Exception as err:
                logger.error(err)
                raise
        else:
            assert different_length_method in ['pad', 'list'], "'different_length_method' can only be 'pad' or 'list'"
            # Apply padding so that all time-series have the same length
            if different_length_method == 'pad':
                logger.info(f'Will pad dataframes with length lower than {self.max_flight_length}')
                idx_flight_lenght_lower_than_max = np.where(self.flights_lengths.flatten() < self.max_flight_length)[0]
                for idx_lower in idx_flight_lenght_lower_than_max:
                    fl_with_diff_length = temp[idx_lower]
                    diff_in_length = self.max_flight_length - len(fl_with_diff_length)
                    padded_df = self.pad_pandas_dataframe(fl_with_diff_length,
                                                          number_of_pads=diff_in_length,
                                                          pad_value=pad_value)
                    temp[idx_lower] = padded_df
                data = np.stack(temp, axis=0)
            elif different_length_method == 'list':
                # Create list of np arrays of dims [L1, D], [L2, D], etc..
                data = []
                for fl in temp:
                    data.append(fl.values)

        self.flight_class = flight_class

        x_train_idx, x_test_idx, y_train, y_test = train_test_split(split_index, flight_class,
                                                                    test_size=test_size,
                                                                    stratify=flight_class)

        # Binary case
        if len(np.unique(y_train)) == 2:
            y_train[np.where(y_train != 0)] = 1
            y_test[np.where(y_test != 0)] = 1

        if not isinstance(data, list):
            self.trainX = data[x_train_idx].reshape(len(x_train_idx),
                                                    self.max_flight_length, -1)
        else:
            # Re-order list
            self.trainX = [data[int(i)] for i in x_train_idx]

        if val_size > 0:
            x_test_idx, x_val_idx, y_test, y_val = train_test_split(x_test_idx, y_test,
                                                                    test_size=val_size,
                                                                    stratify=y_test)
            self.valIndex = x_val_idx
            if not isinstance(data, list):
                self.valX = data[x_val_idx].reshape(len(x_val_idx), self.max_flight_length, -1)
            else:
                self.valX = [data[int(i)] for i in x_val_idx]
            self.valY = y_val
            self.using_val_data = True
        else:
            self.using_val_data = False

        self.trainIndex = x_train_idx
        if not isinstance(data, list):
            self.testX = data[x_test_idx].reshape(len(x_test_idx),
                                                  self.max_flight_length, -1)
        else:
            self.testX = [data[int(i)] for i in x_test_idx]

        self.trainY, self.testY = y_train, y_test
        self.testIndex = x_test_idx

    def normalize_data(self, method: str = 'zscore',
                       avoid_columns: list = []) -> None:

        assert method.lower() in ['zscore', 'minmax'], 'Only "minmax" and "zscore" are available methods'

        self.scaler = StandardScaler() if method == 'standard_scaler' else MinMaxScaler()
        self.data_is_scaled = True

        # Scale data
        if not isinstance(self.trainX, list):
            N_train, L, D = self.trainX.shape
            N_test, _, _ = self.testX.shape
            if self.using_val_data:
                N_val, _, _ = self.valX.shape

            indices_cols_avoid = [self.param_index_dict[param] for param in avoid_columns]
            if len(indices_cols_avoid) > 0:
                # N*L x D
                scaled_x_train = self.scaler.fit_transform(np.delete(self.trainX,
                                                                     indices_cols_avoid,
                                                                     axis=2).reshape(N_train * L,
                                                                                     D - len(indices_cols_avoid)))
                scaled_x_test = self.scaler.transform(np.delete(self.testX,
                                                                indices_cols_avoid,
                                                                axis=2).reshape(N_test * L,
                                                                                D - len(indices_cols_avoid)))
                if self.using_val_data:
                    scaled_x_val = self.scaler.transform(np.delete(self.testX,
                                                                   indices_cols_avoid,
                                                                   axis=2).reshape(N_val * L,
                                                                                   D - len(indices_cols_avoid)))
            else:
                scaled_x_train = self.scaler.fit_transform(self.trainX)
                scaled_x_test = self.scaler.transform(self.testX)
                if self.using_val_data:
                    scaled_x_val = self.scaler.transform(self.valX)

            # Reshape data
            self.scaled_x_train = scaled_x_train.reshape(N_train, L, -1)
            self.scaled_x_test = scaled_x_test.reshape(N_test, L, -1)
            if self.using_val_data:
                self.scaled_x_val = scaled_x_val.reshape(N_val, L, -1)
        else:
            x_train = np.concatenate(self.trainX)
            train_length_each_timeseries = [len(i) for i in self.trainX]
            x_test = np.concatenate(self.testX)
            test_length_each_timeseries = [len(i) for i in self.testX]
            if self.using_val_data:
                x_val = np.concatenate(self.valX)
                val_length_each_timeseries = [len(i) for i in self.valX]

            indices_cols_avoid = [self.param_index_dict[param] for param in avoid_columns]

            if len(indices_cols_avoid) > 0:
                scaled_x_train = self.scaler.fit_transform(np.delete(x_train,
                                                                     indices_cols_avoid,
                                                                     axis=1))
                scaled_x_test = self.scaler.transform(np.delete(x_test,
                                                                indices_cols_avoid,
                                                                axis=1))
                if self.using_val_data:
                    scaled_x_val = self.scaler.transform(np.delete(x_val,
                                                                   indices_cols_avoid,
                                                                   axis=1))
            else:
                scaled_x_train = self.scaler.fit_transform(x_train)
                scaled_x_test = self.scaler.transform(x_test)
                self.scaled_x_test = np.array_split(scaled_x_test, np.cumsum(test_length_each_timeseries))
                if self.using_val_data:
                    scaled_x_val = self.scaler.transform(x_val)

            # Re-create the list of numpy arrays
            self.scaled_x_train = np.array_split(scaled_x_train, np.cumsum(train_length_each_timeseries))
            self.scaled_x_test = np.array_split(scaled_x_test, np.cumsum(test_length_each_timeseries))
            if self.using_val_data:
                self.scaled_x_val = np.array_split(scaled_x_val, np.cumsum(val_length_each_timeseries))

    def retrieve_flight_id_using_flight_idx(self, idx: int, data_set: str = 'train'):

        if not isinstance(self.trainX, list):
            if data_set == 'train':
                return self.trainX[idx, 0, self.param_index_dict[self.flight_id_column]].astype(int)
            elif data_set == 'test':
                return self.testX[idx, 0, self.param_index_dict[self.flight_id_column]].astype(int)
            else:
                return self.valX[idx, 0, self.param_index_dict[self.flight_id_column]].astype(int)
        else:
            if data_set == 'train':
                return self.trainX[idx][0, self.param_index_dict[self.flight_id_column]].astype(int)
            elif data_set == 'test':
                return self.testX[idx][0, self.param_index_dict[self.flight_id_column]].astype(int)
            else:
                return self.valX[idx][0, self.param_index_dict[self.flight_id_column]].astype(int)

    def retrieve_original_flight_data(self, idx: int, data_set: str = 'train',
                                      parameter: str = None):

        parameters_list = [parameter] if parameter is not None else self.parameters_list
        return self.df[self.df[self.flight_id_column] == self.retrieve_flight_id_using_flight_idx(idx,
                                                                                                  data_set)][parameters_list]

    def retrieve_flight_index_using_flight_id(self, flight_id: int,
                                              data_set: str = 'train'):

        if not isinstance(self.trainX, list):
            if data_set == 'train':
                return np.where(self.trainX[:, 0, self.param_index_dict[self.flight_id_column]] == flight_id)[0]
            elif data_set == 'test':
                return np.where(self.testX[:, 0, self.param_index_dict[self.flight_id_column]] == flight_id)[0]
            elif data_set == 'val':
                return np.where(self.valX[:, 0, self.param_index_dict[self.flight_id_column]] == flight_id)[0]
        else:
            if data_set == 'train':
                return [idx for idx in range(len(self.trainX)) if len(np.where(self.trainX[idx][0,
                                                                                                self.param_index_dict[self.flight_id_column]] == flight_id)[0]) > 0][0]
            elif data_set == 'test':
                return [idx for idx in range(len(self.testX)) if len(np.where(self.testX[idx][0,
                                                                                              self.param_index_dict[self.flight_id_column]] == flight_id)[0]) > 0][0]
            elif data_set == 'val':
                return [idx for idx in range(len(self.valX)) if len(np.where(self.valX[idx][0,
                                                                                            self.param_index_dict[self.flight_id_column]] == flight_id)[0]) > 0][0]

    @staticmethod
    def pad_pandas_dataframe(dataframe: pd.DataFrame,
                             pad_value: Union[int, str] = 'max',
                             number_of_pads: int = 1,
                             position: str = 'tailing'):

        if isinstance(pad_value, str):
            assert pad_value.lower() in ['min', 'max', 'mean', 'median'],\
                   '"pads_value" as a str can only be "min", "max", "mean", or "median"'
            if 'max' in pad_value:
                pad_value = dataframe.max().tolist()
                df_temp = pd.DataFrame([pad_value] * number_of_pads,
                                       columns=dataframe.columns)
            elif 'min' in pad_value:
                pad_value = dataframe.min().tolist()
                df_temp = pd.DataFrame([pad_value] * number_of_pads,
                                       columns=dataframe.columns)
            elif 'mean' in pad_value:
                pad_value = dataframe.mean().tolist()
                df_temp = pd.DataFrame([pad_value] * number_of_pads,
                                       columns=dataframe.columns)
            elif 'median' in pad_value:
                pad_value = dataframe.median().tolist()
                df_temp = pd.DataFrame([pad_value] * number_of_pads,
                                       columns=dataframe.columns)
        elif isinstance(pad_value, int):
            df_temp = pd.DataFrame([[pad_value] * dataframe.shape[1]] * number_of_pads,
                                   columns=dataframe.columns)

        if position == 'tailing':
            return dataframe.append(df_temp, ignore_index=True)
        else:
            return df_temp.append(dataframe, ignore_index=True)



from setuptools import setup, find_packages

__version__ = '0.1.0'
__author__ = 'Marc-Henri Bleu-Laine'

setup(
    name='adopt_pytorch',
    author=__author__,
    version=__version__,
    description='Pytorch version of ADOPT algorithm used for precursor mining, '
                'developed by NASA (https://github.com/nasa/ADOPT).'
                'Improvements were made to include time-series of different lengths',
    packages=find_packages(),
    zip_safe=True,
    install_requires=[
        'pandas==1.3.4',
        'numpy==1.21.6',
        'matplotlib==3.5.1',
        'torch==1.12.0',
        'scikit-learn==1.0.2',
        'tqdm==4.64.0',
        # 'torchvision==0.8.2',
        # 'torchaudio==0.7.2'
                    ],
    )



import logging
import config
import numpy as np
import pandas as pd
from typing import List, Union

import torch

from adopt_pytorch.models import ADOPT

logging.basicConfig(level=config.log['feature_ranking'])
logger = logging.getLogger('feature_ranking')


# TODO: Might need to benchmark using torch vs using np for the computations, especially when GPU is not used with torch
def disturb_parameters(clf: ADOPT, input_values: Union[torch.Tensor, np.ndarray] = None,
                       param_index_dict: dict = None, list_parameters: list = None,
                       standard_deviation_scale: List[int] = None,
                       precursor_score_list: list = None) -> list:
    """Perturb each paramters for one flight.

    Args:
        clf (ADOPT): _description_
        input_values (Union[torch.Tensor, np.ndarray], optional): _description_. Defaults to None.
        list_parameters (list, optional): _description_. Defaults to None.
        standard_deviation_scale (List[int], optional): _description_. Defaults to [2, -2].
        precursor_score_list (list, optional): _description_. Defaults to None.
    """

    if standard_deviation_scale is None:
        standard_deviation_scale = config.standard_deviation_scale

    assert standard_deviation_scale[0] > standard_deviation_scale[1], 'Make sure the first index of '\
           'standard_deviation_scale is the upper limit'

    device = clf.device
    if (param_index_dict is not None) and (list_parameters is not None):
        index_parameters = [idx for param, idx in param_index_dict.items() if param in list_parameters]
    elif (list_parameters is not None) and (param_index_dict is None):
        index_parameters = [idx for idx in range(len(clf.parameters_list))]
    else:
        # Perturb all parameters
        index_parameters = [idx for idx in range(len(clf.parameters_list))]
        list_parameters = clf.parameters_list

    # Construct param: modified_precursor_score dict
    feature_precursor_score_dict = {param: {'upper': None, 'lower': None} for param in list_parameters}

    # Time to pertub the parameter and evaluate the precursor score
    if input_values is None:
        # 3D Tensor of shape: N, L, D
        input_values = clf.input_values
    elif torch.is_tensor(input_values):
        input_values = input_values.view(1, input_values.size(0), input_values.size(1))
    else:
        if len(input_values.shape) == 2:
            input_values = input_values.reshape(1, input_values.shape[0], input_values.shape[1])
        input_values = torch.Tensor(input_values)

    # Perform forward pass for regular input
    feature_precursor_score_dict['prediction_likelihood'] = clf(input_values, save_input=False).detach().numpy()
    feature_precursor_score_dict['original_score'] = clf.proba_time.detach().numpy()

    input_values = input_values.to(device)

    # This can be used to combine precursor scores of variable lengths
    if precursor_score_list is None:
        precursor_score_list = []

    for idx_param, param in zip(index_parameters, list_parameters):
        for case_num, case in enumerate(['upper', 'lower']):
            logger.info(f"Case: {case}. Computing pertubation for {param}")
            modified_input_values = input_values.clone()
            feature_input_values = modified_input_values[:, :, idx_param]

            # Compute stardard deviation and apply scale
            standard_deviation_with_scale = torch.std(feature_input_values) * standard_deviation_scale[case_num]

            # Add the std to the input value
            feature_input_values += standard_deviation_with_scale

            # Replace the values in original tensor
            modified_input_values[:, :, idx_param] = feature_input_values

            # Perform forward passes
            _ = clf(modified_input_values, save_input=False).detach().numpy()
            precursor_score = clf.proba_time.detach().numpy()

            feature_precursor_score_dict[param][case] = precursor_score

    precursor_score_list.append(feature_precursor_score_dict)

    return precursor_score_list


def compute_features_precursor_scores(precursor_score_list: List[dict],
                                      standard_deviation_scale: List[int] = None):

    if standard_deviation_scale is None:
        standard_deviation_scale = config.standard_deviation_scale

    param_score_list_of_dict = []
    standard_deviation_scale = standard_deviation_scale + [None]
    param_score_dict = {}
    for precursor_score_dict in precursor_score_list:
        # Shape is N, L, D=1
        original_precursor_score: np.ndarray = precursor_score_dict['original_score'].flatten()
        start, end = identify_start_end_threshold(original_precursor_score)
        for param, value in precursor_score_dict.items():
            if param in ['original_score', 'prediction_likelihood']:
                # skip
                continue
            param_score_dict[param] = {'score': [], 'standard_deviation': []}
            for window_start, window_end in zip(*[start, end]):
                # Usually 1 loop but could be more sometimes...
                mean_original_precursor_score = original_precursor_score[window_start: window_end + 1]
                # Shape is N, L, D=1
                window_upper_precursor_score: np.ndarray = value['upper'].flatten()[window_start: window_end + 1]
                window_lower_precursor_score: np.ndarray = value['lower'].flatten()[window_start: window_end + 1]

                upper_p_diff = round((np.mean(mean_original_precursor_score) -
                                     (np.mean(window_upper_precursor_score))) /
                                     np.mean(mean_original_precursor_score) * 100, 3)
                lower_p_diff = round((np.mean(mean_original_precursor_score) -
                                     (np.mean(window_lower_precursor_score))) /
                                     np.mean(mean_original_precursor_score) * 100, 3)

                param_score_dict[param]['score'].append(min(upper_p_diff, lower_p_diff, 0))
                param_score_dict[param]['standard_deviation'].append(standard_deviation_scale[np.argmin([upper_p_diff,
                                                                                                         lower_p_diff,
                                                                                                         0])])
        param_score_list_of_dict.append(param_score_dict)
    return param_score_list_of_dict


def identify_start_end_threshold(precursor_score: np.ndarray, threshold=0.5):

    if len(precursor_score.shape) > 1:
        precursor_score = precursor_score.flatten()

    flagged_steps = np.where(precursor_score > threshold)[0]
    # Create pandas series
    df = pd.DataFrame(data={'Indices': flagged_steps})
    indices_diff = df.Indices.diff()
    df['window_groups'] = (indices_diff != 1).astype(int).cumsum().values
    windows = df.groupby('window_groups').agg({'Indices': ['min', 'max']}).values
    start = windows.flatten()[0::2]
    end = windows.flatten()[1::2]

    # remove single points
    start_final = []
    end_final = []

    for s, e in zip(*[start, end]):
        if s != e:
            start_final.append(s)
            end_final.append(e+1)
    start_final = np.asarray(start_final)
    end_final = np.asarray(end_final)

    return start_final, end_final

    # flagged_steps = np.asarray([p > threshold for p in precursor_score]).flatten()
    # # Find indices where arrays are not aligned i.e. where the flagging starts and ends
    # flag_steps_indices = np.flatnonzero(flagged_steps[1:] != flagged_steps[:-1])
    # # Start and End indices
    # start = flag_steps_indices[0::2]
    # end = flag_steps_indices[1::2]

    # # last end index would not appear if the score is higher than threshold all the way to the end
    # if len(start) != len(end):
    #     end = np.append(end, len(flagged_steps))

    # return start, end


def rank_features(features_score_list_of_dict: List[dict] = None,
                  precursor_score_list: List[dict] = None,
                  standard_deviation_scale: List[int] = [2, -2]):

    if standard_deviation_scale is None:
        standard_deviation_scale = config.standard_deviation_scale

    sorted_features_list = []
    sorted_scores_list = []
    if features_score_list_of_dict is None and precursor_score_list is None:
        raise '"features_score_list_of_dict" and "precursor_score_list" cannot both be None'

    if precursor_score_list is not None:
        features_score_list_of_dict = compute_features_precursor_scores(precursor_score_list, standard_deviation_scale)

    for features_score_dict in features_score_list_of_dict:
        sorted_parameters_list_of_dict = sorted(features_score_dict.items(),
                                                key=lambda x: min(x[1]['score']))
        sorted_features_list.append([param[0] for param in sorted_parameters_list_of_dict])
        sorted_scores_list.append([abs(min(param[1]['score'])) for param in sorted_parameters_list_of_dict])

    return sorted_features_list, sorted_scores_list



import config
import logging
import numpy as np
import pandas as pd
from typing import List, Union
from matplotlib.axes._axes import Axes
from matplotlib import pyplot as plt

from adopt_pytorch.feature_ranking import identify_start_end_threshold


if 'debug' in config.log['visualization'].lower():
    logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger('visualization')

FIGSIZE = config.FIGSIZE
PLOT_FACE = config.PLOT_FACE
GRID = config.GRID
N_COLUMNS = config.N_COLUMNS
LABEL_SIZE = config.LABEL_SIZE


def plot_precursor_scores(precursor_scores_dict: dict,
                          parameters_list: list,
                          sorted_precursor_params_list: list = None,
                          score_list: list = None,
                          standard_deviation_scale: List[int] = None,
                          thresold: float = 0.5,
                          flight_data: pd.DataFrame = None,
                          filename: str = None, dpi: int = 300):

    if standard_deviation_scale is None:
        standard_deviation_scale = config.standard_deviation_scale

    n_features = len(parameters_list)
    n_rows = 1 if n_features < N_COLUMNS else int(np.ceil((n_features + 1) / N_COLUMNS))

    fig, axes = plt.subplots(n_rows, N_COLUMNS,
                             figsize=(N_COLUMNS * FIGSIZE[0], FIGSIZE[1] * int(n_features / N_COLUMNS + 1)))
    y_label = 'Precursor Score'
    x_label = 'Time'

    current_row = -1
    current_column = 0
    for current_subplot_idx in range(n_features + 1):
        if (current_subplot_idx % N_COLUMNS) == 0:
            current_row += 1
        current_column = current_subplot_idx % N_COLUMNS

        ax = axes[current_row, current_column]
        if current_subplot_idx == 0:
            title = 'Original Precursor Score'
        else:
            title = parameters_list[current_subplot_idx - 1] if sorted_precursor_params_list is\
                    None else sorted_precursor_params_list[current_subplot_idx - 1]
        title_plot = f"{title} ({score_list[current_subplot_idx - 1]} % diff)" \
                     if (score_list is not None) and (current_subplot_idx > 0) else title

        data = precursor_scores_dict['original_score'].flatten()

        ax.plot(data, c='r', label='Original')
        if current_subplot_idx > 0:
            logger.debug(f'Plotting {title}')
            data_up = precursor_scores_dict[title]['upper'].flatten()
            data_low = precursor_scores_dict[title]['lower'].flatten()
            ax.plot(data_up,  label=f"{standard_deviation_scale[0]} σ", c='b')
            ax.plot(data_low, label=f"{standard_deviation_scale[1]} σ", c='g')
            if flight_data is not None:
                ax2 = ax.twinx()
                ax2.plot(flight_data[title].values, label='Feature Values', c='k', ls='dashed')
                ax2 = axis_setup(ax2, y_label='Feature Value')

        start_idx, end_idx = identify_start_end_threshold(data, thresold)
        for start, end in zip(*[start_idx, end_idx]):
            ax.axvspan(start, end,
                       alpha=0.25,
                       color='gray')

        ax = axis_setup(ax,
                        title=title_plot,
                        y_label=y_label,
                        x_label=x_label,
                        legend=True)

    fig.tight_layout(pad=0.95)

    if filename is not None:
        fig.save(filename, dpi=dpi)


def axis_setup(ax: Axes, title: str = None,
               y_label: str = None, x_label: str = None,
               tick_axis: str = 'both',
               which_tick_axis: str = 'major',
               legend: Union[bool, str] = None):

    ax.set_facecolor(PLOT_FACE)
    ax.grid(GRID)
    if y_label is not None:
        ax.set_ylabel(y_label, fontsize=LABEL_SIZE)
    if x_label is not None:
        ax.set_xlabel(x_label, fontsize=LABEL_SIZE)
    ax.tick_params(axis=tick_axis, which=which_tick_axis, labelsize=LABEL_SIZE)
    if title is not None:
        ax.set_title(title)
    if isinstance(legend, bool):
        ax.legend()
    elif isinstance(legend, str):
        ax.legend(legend)

    return ax



import config
import torch
import logging
import warnings
import numpy as np
from torch import nn
from tqdm import tqdm
from sklearn.metrics import f1_score, balanced_accuracy_score
from torch.utils.data import DataLoader, WeightedRandomSampler

from adopt_pytorch.data import SimpleData

logging.basicConfig(level=config.log['models'])
logger = logging.getLogger('models')
torch.manual_seed(config.pytorch_seed)
np.random.seed(config.numpy_seed)


class ADOPT(nn.Module):
    '''Pytorch version of ADOPT. This version is able to use variable flight lengths'''
    def __init__(self, n_features, parameters_list,
                 hidden_size=5, device='cpu',
                 max_seq_len=None):
        nn.Module.__init__(self)

        self.n_features = n_features
        self.parameters_list = parameters_list
        self.hidden_size = hidden_size
        self.device = device
        self.max_seq_len = max_seq_len
        self.input_values = None

        self.gru = nn.GRU(input_size=self.n_features,
                          hidden_size=self.hidden_size,
                          batch_first=True)

        self.tanh_ = nn.Tanh()

        self.dense1 = nn.Sequential(nn.Linear(in_features=self.hidden_size,
                                    out_features=500),
                                    nn.Tanh())

        self.dense2 = nn.Sequential(nn.Linear(in_features=500,
                                    out_features=1),
                                    nn.Sigmoid())

    def forward(self, x, save_input=True):
        if not torch.is_tensor(x):
            x = torch.Tensor(x).to(self.device)
        else:
            x = x.to(self.device)

        if len(x.size()) == 2:
            x = x.view(1, x.size(0), x.size(1))

        # Save the passed input
        if save_input:
            self.input_values = x.detach().clone()
        # Make sure that the precursor probability over time is specific to current batch
        batch_size = x.size(0)
        n_timesteps = x.size(1)
        self.precursor_proba = torch.zeros((batch_size,
                                            n_timesteps,
                                            self.n_features)).to(self.device)

        self.proba_time = torch.zeros((batch_size,
                                       n_timesteps)).to(self.device)

        # Stateless
        h_0 = torch.zeros(1, batch_size, self.hidden_size).to(self.device)
        gru_out, h_out = self.gru(x, h_0)
        out1 = self.dense1(self.tanh_(gru_out))
        self.proba_time = self.dense2(out1)
        # Take the max proba across time
        final_out = torch.max(self.proba_time, axis=1)[0]

        return final_out.view(-1)


def train_binary_model(clf, X_train, y_train,
                       X_val=None, y_val=None, model_out_cpu: bool = True,
                       threshold: float = 0.5, batch_size: int = None, weight=None, optimizer: str = 'adam',
                       l2: float = 0, learning_rate: float = 0.01, n_epochs: int = 100,
                       use_stratified_batch_size: bool = True, print_every_epochs: int = 10,
                       verbose: int = 0, ignore_warnings=False):

    if ignore_warnings:
        warnings.filterwarnings('ignore')

    # Set the model to the relevant device
    clf.to(clf.device)

    # Add weight to each example, kinda useful for imbalanced dataset at times
    if weight is not None:
        weight = torch.Tensor(weight).to(clf.device)
    if len(np.unique(y_train)) == 2:
        criterion = nn.BCELoss(weight=weight)
    else:
        raise NotImplementedError

    # Setup optimizer
    if optimizer == "adam":
        optimizer = torch.optim.Adam(clf.parameters(),
                                     lr=learning_rate, weight_decay=l2)
    else:
        optimizer = torch.optim.SGD(clf.parameters(),
                                    lr=learning_rate, weight_decay=l2)

    # Init loss history and balanced accuracy
    hist = np.zeros(n_epochs)
    val_hist = np.zeros(n_epochs)
    b_acc = np.zeros(n_epochs)
    val_b_acc = np.zeros(n_epochs)
    f1 = np.zeros(n_epochs)
    val_f1 = np.zeros(n_epochs)

    # Conversion to tensors
    if not torch.is_tensor(X_train):
        if isinstance(X_train, list):
            X_train = [torch.Tensor(x) for x in X_train]
        else:
            X_train = torch.Tensor(X_train)
    if not torch.is_tensor(y_train):
        y_train = torch.Tensor(y_train.flatten())

    if X_val is not None and y_val is not None:
        if not torch.is_tensor(X_val):
            if isinstance(X_val, list):
                X_val = [torch.Tensor(x) for x in X_val]
            else:
                X_val = torch.Tensor(X_val)
        if not torch.is_tensor(y_val):
            y_val = torch.Tensor(y_val)
        data_val = SimpleData(X_val, y_val)

    if batch_size is None:
        batch_size = X_train.size(0)
        logger.warn("Setting the batch size = full training set could overwhelm GPU memory")
    if isinstance(X_train, list):
        batch_size = 1
        logger.warn('Beware, changing the batch size to 1 and training may be slow!')

    data_train = SimpleData(X_train, y_train)
    if use_stratified_batch_size is False:
        logger.info("Mini-batch strategy: Random sampling")
        dataloader_train = DataLoader(data_train, batch_size=batch_size, shuffle=True)
    else:
        logger.info("Mini-batch strategy: Stratified")
        # get class counts
        weights = []
        for label in torch.unique(y_train):
            count = len(torch.where(y_train == label)[0])
            weights.append(1 / count)
        weights = torch.tensor(weights).to(clf.device)
        samples_weights = weights[y_train.type(torch.LongTensor).to(clf.device)]
        sampler = WeightedRandomSampler(samples_weights, len(samples_weights), replacement=True)
        dataloader_train = DataLoader(data_train, batch_size=batch_size, sampler=sampler)

    if X_val is not None:
        dataloader_val = DataLoader(data_val, batch_size=batch_size, shuffle=False)

    # Train the model
    try:
        for epoch in tqdm(range(n_epochs)):
            batch_acc = []
            batch_val_acc = []
            batch_f1 = []
            batch_val_f1 = []

            for iteration, (batch_x, batch_y) in enumerate(dataloader_train):
                # Get batch
                batch_x, batch_y = batch_x.to(clf.device), batch_y.to(clf.device)
                # Reset gradient to zero
                optimizer.zero_grad()
                if (epoch == 0) and iteration == 0:
                    # TODO: Change this
                    for c in torch.unique(y_train):
                        logger.info(f"Proportion Class {c}: {batch_y[batch_y==c].shape[0]/len(batch_y)}")

                # Forward pass
                outputs = clf(batch_x)
                # Get the loss
                loss = criterion(outputs.flatten(), batch_y.view(-1).flatten())

                hist[epoch] = loss.item()

                if "cuda" in clf.device:
                    temp_outpouts = (outputs.cpu().detach().numpy() > threshold).astype(int)
                    y_batch = batch_y.view(-1).cpu().detach().numpy()
                    b_acc[epoch] = balanced_accuracy_score(y_batch, temp_outpouts)
                else:
                    temp_outpouts = (outputs.detach().numpy() > threshold).astype(int)
                    y_batch = batch_y.view(-1).detach().numpy()
                    b_acc[epoch] = balanced_accuracy_score(y_batch, temp_outpouts)

                batch_acc.append(b_acc[epoch])
                batch_f1.append(f1_score(y_batch, temp_outpouts, average='binary'))

                # Backprop and perform optimisation
                loss.backward()
                optimizer.step()

                if X_val is not None and y_val is not None:
                    # Do not compute gradient on validation
                    with torch.no_grad():
                        mini_loss = []
                        for batch_X_val, batch_y_val in dataloader_val:
                            batch_X_val, batch_y_val = batch_X_val.to(clf.device), batch_y_val.to(clf.device)
                            valYhat = clf(batch_X_val)

                            val_loss = criterion(valYhat, batch_y_val.flatten())

                            mini_loss.append(val_loss.item())

                            if "cuda" in clf.device:
                                temp_out_y = (valYhat.cpu().detach().numpy() > threshold).astype(int)
                                y_val_batch = batch_y_val.view(-1).cpu().detach().numpy()
                                val_b_acc[epoch] = balanced_accuracy_score(y_val_batch,
                                                                           temp_out_y)
                            else:
                                temp_out_y = (valYhat.detach().numpy() > threshold).astype(int)
                                y_val_batch = batch_y_val.view(-1).detach().numpy()
                                val_b_acc[epoch] = balanced_accuracy_score(y_val_batch,
                                                                           temp_out_y)
                            batch_val_acc.append(val_b_acc[epoch])
                            batch_val_f1.append(f1_score(y_val_batch, temp_out_y, average='binary'))
                        val_hist[epoch] = np.mean(mini_loss)

                if verbose == 1:
                    if epoch % print_every_epochs == 0:
                        logger.info("\nEpoch: %d, loss: %1.5f" % (epoch, loss.item()))
                        logger.info("Epoch: %d, b_acc: %1.5f" % (epoch, b_acc[epoch]))
                        logger.info("Epoch: %d, f1 (binary): %1.5f" % (epoch, f1_score(y_batch,
                                                                                       temp_outpouts,
                                                                                       average='binary')))
                        if X_val is not None:
                            logger.info("Epoch: %d, val_loss: %1.5f" % (epoch, val_hist[epoch]))
                            logger.info("Epoch: %d, val_b_acc: %1.5f" % (epoch, val_b_acc[epoch]))
                            logger.info("Epoch: %d, val_f1 (binary): %1.5f\n" % (epoch, f1_score(y_val_batch,
                                                                                                 temp_out_y,
                                                                                                 average='binary')))

            b_acc[epoch] = np.mean(batch_acc)
            val_b_acc[epoch] = np.mean(batch_val_acc)
            f1[epoch] = np.mean(batch_f1)
            val_f1[epoch] = np.mean(batch_val_f1)
    except KeyboardInterrupt:
        logger.info("Training was interuppted!")
        if model_out_cpu and clf.device != "cpu":
            # Puts back model on cpu
            clf.cpu()
            clf.device = "cpu"

        # Eval/testing mode
        clf.eval()

        # Keep training data inside of model class
        clf.x_train = X_train
        clf.x_test = X_val

        return clf, hist, val_hist, (b_acc, val_b_acc, f1, val_f1)

    except Exception as E:
        logger.warn(E)
        raise

    if model_out_cpu and clf.device != "cpu":
        # Puts back model on cpu
        clf.cpu()
        clf.device = "cpu"

    # Eval/testing mode
    clf.eval()

    clf.x_train = X_train
    clf.x_test = X_val

    return clf, hist, val_hist, (b_acc, val_b_acc, f1, val_f1)
